/*
 * cocengine.c
 *
 * Copyright (c) 1999,2000,2001 Hajime Mochizuki <motizuki@jaist.ac.jp>
 *   Natural Language Processing Laboratory,
 *   Japan Advanced Institute of Science and Technology.
 *
 *
 * Use of this software is granted only for RESEARCH PURPOSES. 
 * Redistribution is granted as long as all copyright notices are retained.
 * Redistribution with modification is allowed provided that you make your 
 * modified version obviously distinguishable from the original one.  
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES ARE DISCLAIMED.
 *
 * geta б
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <geta/config.h>
#include <geta/wam.h>
#include "coscr.h"

/* #include "cmn.h"*/
#include "mrpfil.h"
#include "cochainer.h"
/*#include "cooption.h"*/

/* for CHAR_BIT */
#if defined(BSD44) || defined(DIGITAL)
#include <machine/limits.h>
#else
#include <limits.h>
#endif

extern CHAFLT_OUT *chaflt_out; /* mrpflt.c */
extern int chaflt_out_num;     /* mrpflt.c */
extern int number_of_words;    /* mrpflt.c */
extern int co_scr_num;         /* coscr.c */
extern double threshold;       /* coscr.c */
extern struct _copair copair;  /* coscr.c */

struct xrlist *xrhashtab[XRHASHSIZE];
struct colist *cohashtab[COHASHSIZE];

int mrp_type;

extern XRinfo*     get_xrinfo ();
extern double      get_max_score ();
extern int         calc_pair_coscr();

struct tcblist *top_chain;
struct tcblist *btm_chain;
struct tcoscr  maxtco;

void merror (f)
     char *f;
{
	fprintf (stderr, "cannot malloc for %s\n", f);
	exit(1);
}

unsigned xrhash (s)
    char *s;
{
	unsigned xrhashval;

	if (strlen(s)==0) {
		fprintf (stderr, "!!!!\n");
	}
	for (xrhashval=0; *s!='\0'; s++) {
		xrhashval = (( xrhashval<<CHAR_BIT) + *s) % XRHASHSIZE;
	}
	return xrhashval;
}

struct xrlist *xrhash_lookup (s)
    char *s;
{
	struct xrlist *np;
	np = NULL;

	for (np = *(xrhashtab+(xrhash (s))); np!=NULL; np=np->next)
	    if (strcmp(s, np->word)==0)
		return np;        /* Ĥä */
	return NULL;              /* Ĥʤ */
}

struct xrlist *xrhash_add (word, w)
     char *word;
     WAM   *w;
{
	struct xrlist       *np, *p;
	unsigned xrhashval;


	xrhashval = xrhash(word);
	if (*(xrhashtab+xrhashval)!=NULL) { /* ֤ꤢ */
		for (p = *(xrhashtab+xrhashval); p!=NULL; p=p->next) {
			if (strcmp(word, p->word)==0) { /* ¸ߤƤ */
				return p;
			}
		}
	}
	if ((np = (struct xrlist *)calloc((size_t)1, (size_t)sizeof(*np))) == NULL)
	    merror ("xrhashtab");
	np->word = strdup(word);
	np->xri  = get_xrinfo (word, w);
	if (np->xri->doc==NULL) {
		fprintf (stderr, "%s\n", word);
	}
	np->next = *(xrhashtab+xrhashval);
	*(xrhashtab+xrhashval) = np;	
	return np;
}	

void xrhash_update (words, w)
    CHAFLT_OUT *words;
    WAM   *w;
{
	int i;
	struct xrlist *exist;

	exist=NULL;
	for (i=0; i<chaflt_out_num; i++) {
		exist = xrhash_add (words[i].term, w); /* ʤɲä */
	}
}

unsigned cohash (s)
    char *s;
{
	unsigned cohashval;

	if (strlen(s)==0) {
		fprintf (stderr, "!!!!\n");
	}
	for (cohashval=0; *s!='\0'; s++) {
		cohashval = (( cohashval<<CHAR_BIT) + *s) % COHASHSIZE;
	}
	return cohashval;
}

void co_add (exist1, exist2, calc_mode, N) 
     struct xrlist *exist1;
     struct xrlist *exist2;
     int    calc_mode, N;
{
	int  dam, id1, id2;
	int  key;
/*	char   key[100];*/
	struct colist *np, *p, *q;
	unsigned cohashval;

	key=0;
	np=p=q=NULL;
	
	dam=0;
	if (exist1==NULL || exist2==NULL) {
		copair.scr = (double)0;
		copair.df[0] = copair.df[1] = copair.df[2] = exist1->xri->nrecords;
		return;
	}
	if (exist1->xri->id==-1 || exist2->xri->id==-1) {
		copair.scr = (double)0;
		copair.df[0] = copair.df[1] = copair.df[2] = exist1->xri->nrecords;
		return;
	}
	if (exist1->xri->id < exist2->xri->id) {
		id1 = exist1->xri->id;
		id2 = exist2->xri->id;
		key = id1 + (id2*10);
/*	    sprintf (key, "%f", 
		     (double)(exist1->xri->id * JOUIVAL) + (double)exist2->xri->id);
*/	
	}
	else if (exist1->xri->id > exist2->xri->id) {
		id1 = exist2->xri->id;
		id2 = exist1->xri->id;
	    key = id1 + (id2*10);
/*	    sprintf (key, "%f",
		     (double)(exist2->xri->id * JOUIVAL) + (double)exist1->xri->id);*/
	}
	else {
		copair.scr   = max_score;
		copair.df[0] = copair.df[1] = copair.df[2] = exist1->xri->nrecords;
		return;
	}
/*	cohashval = cohash (key);*/
	cohashval = key % COHASHSIZE;

	if (*(cohashtab+cohashval)!=NULL) {
		for (p = *(cohashtab+cohashval); p!=NULL; p=p->next) {

/*			if (key == p->key) {*/
/*			if (strcmp(key, p->key)==0) {*/
			if (id1==p->id1 && id2==p->id2) {
				copair.scr = p->scr; /* Ĥä */
				return;
			}
		}
	}
	if ((np = (struct colist *)calloc((size_t)1, sizeof(struct colist)))==NULL)
	    merror ("cohashtab");
/*	np->key  = key;*/
/*
	if ((np->key = (char *)calloc((size_t)(strlen(key)+1), sizeof(char)))==NULL)
	    merror ("np->key");
	strcpy(np->key, key);
*/
/*	np->key = strdup(key);*/
	np->id1 = id1;
	np->id2 = id2;

	dam = calc_pair_coscr (exist1->xri, exist2->xri, calc_mode, N);
	np->scr = copair.scr;

	np->next = *(cohashtab+cohashval);
	*(cohashtab+cohashval) = np;
	return;
}

CLUSTER *addword (ws, sn, wn, nm)
      char *ws;
      int   sn, wn, nm;
{
	CHAFLT_OUT *ca;
	CLUSTER *sc;

	ca = NULL;
	sc = NULL;

	if ((sc=(CLUSTER *)calloc((size_t)1, sizeof(CLUSTER)))==NULL)
	    merror ("Cluster");
	if ((ca=(CHAFLT_OUT *)calloc((size_t)1, sizeof(CHAFLT_OUT)))==NULL)
	    merror ("Chaflt_out");
	ca->term = strdup(ws);
	ca->snum = sn;       ca->wnum = wn;     ca->number = nm;
	sc->elem = ca;       sc->next = NULL;
	return sc;
}

void scrcpy  (scr, i, j, a)
     CO_SCR scr[];
     int    i, j, a;
{
	scr[a].numbers[0] = i;	     scr[a].numbers[1] = j;
	scr[a].df[0] = copair.df[0]; scr[a].df[1] = copair.df[1];
	scr[a].df[2] = copair.df[2]; scr[a].scr   = copair.scr;
}

struct clusterbase *sentence_level_chaining (terms)
    CHAFLT_OUT *terms;
{
	int i, j, a, w1, w2, w3, w4, flg;
	struct xrlist       *exist1, *exist2;
	struct clusterbase  *bp, *start_bp, *pbp, *xp, *nbp,*basep;
	struct cbh          *cbhs;
	CLUSTER             *sc, *sc_forward, *xsc, *xsc_tmp;
	CHAFLT_OUT          *ca;
	CO_SCR              *scr;

	a = flg = w1 = w2 = w3 = w4 = 0;
	exist1 = exist2 = NULL;
	bp = start_bp = pbp = xp = nbp = basep = NULL;
	cbhs = NULL;
	ca = NULL;
	scr = NULL;

	a = chaflt_out_num * (chaflt_out_num-1) / 2;
	if (a==0) { /* ñ1Ĥλ ¨return ǤϤޤΤ */
		/* sc = CLUSTER */
		if ((sc=calloc((size_t)1, sizeof(*sc)))==NULL) {
			merror("sc");
		}
		sc->elem = &terms[0];
		sc->next = NULL;

		/* bp = cluster base pointer */
		if ((bp=calloc((size_t)1, sizeof(*bp)))==NULL) {
			merror ("clusterbase");
		}
		bp->up   = NULL;
		bp->top  = sc;
		bp->down = NULL;
		return bp;
	}

	/** terms 2İʾ **/

 	if ((scr=calloc((size_t)a, sizeof(*scr)))==NULL) {
		merror ("scr (CO_SCR)");
	}
	co_scr_num = a;
	a = 0;

	/* cluster base hash(Ϥ)κ */	   
	if ((cbhs=calloc((size_t)(chaflt_out_num+1), sizeof(*cbhs)))==NULL) {
		merror ("cbh");
	}
	w1 = terms[0].wnum;
/*	w1 = terms[0].wnum-1;*/

	/* sc = CLUSTER */
	if ((sc=calloc((size_t)1, sizeof(*sc)))==NULL) {
		merror("sc");
	}
	sc->elem = &terms[0];
	sc->next = NULL;

	/* bp = cluster base pointer */
	if ((bp=calloc((size_t)1, sizeof(*bp)))==NULL) {
		merror ("clusterbase");
	}
	bp->up   = NULL;
	bp->top  = sc;
	bp->down = NULL;
	start_bp = bp;
	pbp = start_bp;

	cbhs[w1].bs = bp;  /* wordnumber */
	cbhs[w1].pos= 0;   /* pointer order*/

	/* η׻ 򤷤Ĥ cbhs, bp(base pointer)ν */
	for (i=0; i<chaflt_out_num-1; i++) {

		/* i+1ܤξ, cbhs, bp, sc */
		if ((sc=calloc((size_t)1, sizeof(*sc)))==NULL) {
			merror("sc");
		}
		sc->next = NULL;
		sc->elem =&terms[(i+1)];
		fflush (stderr);
		if ((bp=calloc((size_t)1, sizeof(*bp)))==NULL) {
			merror("clbase");
		}
		w1 = terms[(i+1)].wnum;
/*		w1 = terms[(i+1)].wnum-1;*/
		bp->up   = pbp;
		bp->top  = sc;    /* äbpˡCLUSTERդ */
		bp->down = NULL;

		cbhs[w1].bs = bp;    /* wordnumber */
		cbhs[w1].pos= (i+1); /* pointer order*/

		pbp->down = bp;
		pbp       = bp;

		exist1 = xrhash_lookup (terms[i].term);
		for (j=i+1; j<chaflt_out_num; j++) {
			exist2 = xrhash_lookup (terms[j].term);
			co_add (exist1, exist2, calc_mode, N); /* scr׻ */
			scrcpy (scr, i, j, a);
			a++;
		}
	}
	co_scr_sort (0, co_scr_num-1, scr); /* scrˤ¤ٴ */

	for (i=0; i<co_scr_num; i++) {
		if (scr[i].scr < threshold) { /* ͤʤ罪 */
			i = co_scr_num;
			break;
		}
		w1 = terms[scr[i].numbers[0]].wnum;
		w2 = terms[scr[i].numbers[1]].wnum;

/*		w1 = terms[scr[i].numbers[0]].wnum-1;
		w2 = terms[scr[i].numbers[1]].wnum-1;
*/
		if (cbhs[w1].pos == cbhs[w2].pos) { /* Ʊ饹ˤ */
			continue;
		}
		else if (cbhs[w1].pos < cbhs[w2].pos) {
			bp = cbhs[w1].bs;
			xp = cbhs[w2].bs;  /* ȤǾä */
			w3 = w1;
		} else {
			bp = cbhs[w2].bs;
			xp = cbhs[w1].bs;
			w3 = w2;
		}
		for (sc=xp->top; sc!=NULL; sc=sc->next) {
			w4 = sc->elem->wnum;          /*baseν񴹤 */
			cbhs[w4].pos = cbhs[w3].pos;
			cbhs[w4].bs  = cbhs[w3].bs;
		}
		/***** 饹Υޡ (ñνи˥ꥹȤѤ) *****/
		xsc = xp->top;  /* xscϤ줫Ĥʤ cluster1 */
		sc  = bp->top;  /* scϤĤʤ cluster1 */
		sc_forward = NULL;
		while (xsc!=NULL) {
			flg = 0;
			while (sc!=NULL) {
/*if(sc->elem) fprintf(stderr, "."); else fprintf(stderr, "!");*/
				if (sc->elem->wnum > xsc->elem->wnum) {
					break;
				}
				sc_forward = sc;
				sc  = sc->next;
			}
			sc_forward->next = xsc;  /* scp -> sc  scp-> xsc  */
			xsc_tmp    = xsc;        /* ¸ */
			xsc        = xsc->next;  /* xscν */
			xsc_tmp->next = sc;      /* scp -> xsc -> sc  δ  */
			sc_forward = xsc_tmp;       /* ­Ȥ鸫Ф */
		}
		/************************************************** ޡޤ */
		pbp       = xp->up;
		pbp->down = xp->down;
		if (pbp->down!=NULL) {
			pbp->down->up = pbp;
		}
/*		nbp       = xp->down;
		if (nbp!=NULL) nbp->up = pbp;
*/
/*#ifdef NIS030698_XXX*/
/*		free (xp); xp=NULL;*/
/*#endif*/
	}
	basep = start_bp;

	if (debug==1) {
		for (bp=start_bp; bp!=NULL; bp=bp->down) {
			for (sc=bp->top; sc!=NULL; sc=sc->next) {
				fprintf (stderr, "%s ", 
					 sc->elem->term);
				fflush (stderr);
				fprintf (stderr, "%d - ", sc->elem->number);
			}
			fprintf (stderr, "\n");
			fflush (stderr);
		}
		fprintf (stderr, "\n");
		fflush (stderr);
	}
	free (cbhs);
	cbhs=NULL;
	if (co_scr_num>0) { free (scr); scr = NULL; }
	return basep;
}

struct clusterbase *ma (tp)
     struct clusterbase *tp;
{
	if ((tp=(struct clusterbase *)calloc
	     ((size_t)1, sizeof(struct clusterbase)))==NULL)
	    merror ("tp");
	return tp;
}


void print_chains (sw,wsw,comp)
     int sw,wsw,comp;
{
	struct tcblist *a;
	struct tcelist *b;
	CLUSTER *c;
	int     i;

	i=0;  a = NULL; b = NULL; c = NULL;

	fprintf (stderr, "top_chain = %d, btm_chain = %d\n",
		top_chain, btm_chain);

	for (a=top_chain; a!=NULL; a=a->down) {
		if (sw==1) printf ("%2d ", i);
		for (b=a->top; b!=NULL; b=b->next) {
			for (c=b->top; c!=NULL; c=c->next) {
				if (sw==1) 
				    printf ("[%s %d]-",
					    c->elem->term, c->elem->number);
				else if (sw==2) {
					if (wsw==0) 

						printf ("%d %d %d ",
							c->elem->snum,
							c->elem->wnum,
							c->elem->number);
					else
						printf ("%d %d %d ",
							c->elem->snum,
							c->elem->realwn,
							c->elem->number);
					if (comp==0) 
						printf ("%s %s %s\n", 
							c->elem->term,
							c->elem->pp, c->elem->ppbase);
					else 
						printf ("* %s %s %s\n", 
							c->elem->term,
							c->elem->pp, c->elem->ppbase);
				}
			}
		}
		printf ("\n");
		i++;
	}
}

void sort_tcoscr (left, right, a)
       int    left, right;
       struct tcoscr a[];
{
	int    i,  j;
	double x;
	struct tcoscr s;

	i = j = 0; x = (double)0;

	if (left<right) {
		x = a[((left+right)/2)].scr;
		i = left-1; j = right+1;
		while (1) {
			while (a[++i].scr > x);
			while (a[--j].scr < x);
			if  (i>=j)  break;
			s    = a[i];
			a[i] = a[j];
			a[j] = s;
		}
		sort_tcoscr (left, i-1,   a);
		sort_tcoscr (j+1,  right, a);
	}
}

int get_max_co (a, p) /* 祹׻ */
	struct tcblist     *a;
	struct clusterbase *p;
{
	int ct;
	struct xrlist  *exist1, *exist2;
	struct tcelist *te;
	CLUSTER   *tc, *sc;

	exist1 = exist2 = NULL;
	te = NULL;
	tc = sc = NULL;
	ct = 0;
	maxtco.scr   = (double)0;
	maxtco.tcb   = a;
	maxtco.sctop = p;
	for (sc=p->top; sc!=NULL; sc=sc->next) {
/*		printf ("%s\n", sc->elem->term);*/
		exist1 = xrhash_lookup (sc->elem->term); /* seg fault*/
		for (te=a->top; te!=NULL; te=te->next) {
			for (tc=te->top; tc!=NULL; tc=tc->next) {
				exist2 = xrhash_lookup (tc->elem->term);
				co_add (exist1, exist2, calc_mode, N);
				ct++;
				if (copair.scr == max_score) {
					maxtco.scr = copair.scr;
					return ct;
				} else if (maxtco.scr<copair.scr) {
					maxtco.scr = copair.scr;
				}
			}
			/* äȥϸʤ */
			if (te==a->old_last) /*return maxtco; */
			    return ct;
		}
	}
	return ct;
}

void text_level_chaining (start_bp)
     struct clusterbase  *start_bp;
{
	struct clusterbase *cbp_p, *cbp_q;
	struct tcblist     *tcb_a, *tcb_b;
	struct tcelist     *tce_e, *tce_f;
	CLUSTER            *tc, *sc;
	struct tcoscr      *co_scr;/*, *tmp;*/
	int                 i, j, ct;

	i = j = ct =0;

	if (top_chain == NULL) { /* ϰȯ */
		for (cbp_p=start_bp; cbp_p!=NULL; cbp_p=cbp_p->down) {
			if ((tcb_a=calloc((size_t)1, sizeof(*tcb_a)))==NULL)
			    merror ("tcblist no a.");
			if ((tce_e=calloc((size_t)1, sizeof(*tce_e)))==NULL)
			    merror ("tcelist no e.");
			tce_e->next = NULL;
			tce_e->top  = cbp_p->top;
			tcb_a->up   = NULL;
			tcb_a->top  = tce_e;
			tcb_a->last = tce_e;
			tcb_a->old_last = tce_e;
			tcb_a->check = 1;
			tcb_a->down = top_chain;
			if (top_chain!=NULL) {
				top_chain->up = tcb_a;
			}
			else {
				btm_chain     = tcb_a;
			}
			top_chain = tcb_a;
		}
	} else {
		cbp_p=start_bp;
		if ((co_scr =calloc((size_t)2000, sizeof(struct tcoscr)))==NULL)
		    merror ("tcoscr no coscr");

		while (cbp_p!=NULL) {
			i = 0;
/*			fprintf (stderr, "calc get_max_co s in\n");*/
			for (tcb_a=top_chain; tcb_a!=NULL; tcb_a=tcb_a->down) {
				if (tcb_a->check == 0) continue;
				ct+=get_max_co (tcb_a, cbp_p);
if(i>=2000) fprintf(stderr, "!!!\n");
/*				co_scr[i].scr   = tmp->scr;*/
				co_scr[i].scr   = maxtco.scr;
				co_scr[i].tcb   = tcb_a;
/*				co_scr[i].sctop = p->top;*/
				co_scr[i].sctop = cbp_p;
				i++;
			}
/*			fprintf (stderr, "calc get_max_co s out %d\n", ct);*/
			sort_tcoscr (0, i-1, co_scr);
			j=0;
			if (co_scr[j].scr >= threshold) { /* ȤˤζΤ
							 * äĤ */
				tcb_a = co_scr[j].tcb;
				if ((tce_e=calloc((size_t)1, sizeof(*tce_e)))==NULL)
				    merror ("tcelist no e.");
				tce_e->next = NULL;
				tce_e->top  = cbp_p->top;
if(j>=2000) fprintf(stderr, "!!!\n");
				tce_f = tcb_a->last;
				tce_f->next = tce_e; /* last­ */
				tcb_a->last = tce_e; /* last򹹿   */

				if (tcb_a!=top_chain) { /* ȥåפ˰ư */
					top_chain->up = tcb_a;
					if (tcb_a==btm_chain) {
						btm_chain   = tcb_a->up;
					}
					else {
						tcb_a->down->up = tcb_a->up;
					}
					tcb_a->up->down = tcb_a->down;
					tcb_a->down     = top_chain;
					top_chain = tcb_a;
				}
			} else { /* ɤˤ⤯äĤʤΤǡ */
				if ((tcb_a=calloc((size_t)1, sizeof(*tcb_a)))==NULL)
				    merror ("tcblist no a.");
				if ((tce_e=calloc((size_t)1, sizeof(*tce_e)))==NULL)
				    merror ("tcelist no e.");
				tce_e->next = NULL;
				tce_e->top  = cbp_p->top;
				tcb_a->up   = NULL;
				tcb_a->top  = tce_e;
				tcb_a->last = tce_e;
				tcb_a->old_last = tce_e;
				tcb_a->down = top_chain;
				top_chain->up = tcb_a;
				top_chain = tcb_a;
			}
			cbp_q = cbp_p;
			cbp_p = cbp_p->down;
/* check check!!
			free (cbp_q); cbp_q = NULL;
 */
		}
		free (co_scr);
 	}
	for (tcb_a=top_chain; tcb_a!=NULL; tcb_a=tcb_a->down) {
		tcb_a->check = 1;
		tcb_a->old_last = tcb_a->last;
	}
	if (debug==1)
	    fprintf (stderr, "co-occurence pair = %d\n", ct);

#ifdef TEST3
	print_chains(1);
#endif
}
