/*
 * mkidx.c
 *
 * Copyright (c) 1997-2000,2001 Hajime Mochizuki <motizuki@jaist.ac.jp>
 *   Natural Language Processing Laboratory,
 *   Japan Advanced Institute of Science and Technology.
 *
 *
 * Use of this software is granted only for RESEARCH PURPOSES. 
 * Redistribution is granted as long as all copyright notices are retained.
 * Redistribution with modification is allowed provided that you make your 
 * modified version obviously distinguishable from the original one.  
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES ARE DISCLAIMED.
 *       
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

off_t pointer;
int   dic;
int   field;
int debug;
static char *pname;
char *fn;

static help ()
{
	fprintf (stderr, "mkidx ver.0.2\n");
	fprintf (stderr, "Usage: %s < dicfile\n", pname);
	fprintf (stderr, "\t[-F #]   ϥե(dic.dat)\n");
	fprintf (stderr, "\t[-dic #] μ\n");
	fprintf (stderr, "\t\t0 :chainѤbgh\n");
	fprintf (stderr, "\t\t1 :chainѤkadokawa\n");
	fprintf (stderr, "\t\t2 :neroѤbgh\n");
	fprintf (stderr, "\t\t3 :neroѤkadokawa\n");
	fprintf (stderr, "\t\t5 :chainѤntt\n");
	fprintf (stderr, "\t[-W|R|X] W:Ф,R:ɤ,X:ξ indexing\n");
	fprintf (stderr, "\t[-d]     debug mode\n");
	fprintf (stderr, "\t[-h]     help\n");
	exit(1);
}

static optchk (argc, argv) 
    int   argc;
    char *argv[];
{
	pname = argv[0];
	argc--, argv++;
	dic = 1;
	fn = "dic.dat";
	field=2;
	debug=0;
	while (argc>0 && *argv[0]=='-') {
		if (!strcmp(argv[0], "-h")) {
			help();
		} else if (!strcmp(argv[0], "-dic")) {
			argc--, argv++;
			dic = atoi(argv[0]);
		} else if (!strcmp(argv[0], "-F")) {
			argc--, argv++;
			fn=argv[0];
		} else if (!strcmp(argv[0], "-W")) {
			field=0;
		} else if (!strcmp(argv[0], "-R")) {
			field=1;			
		} else if (!strcmp(argv[0], "-X")) {
			field=2;
		} else if (!strcmp(argv[0], "-d")) {
			debug=1;
		} else {
			fprintf (stderr, "Invalid arg: %s\n", argv[0]);
			help();
		}
		argc--, argv++;
	}
}

void mkdata (s, k1, k2, i1, i2, wf, i, sw)
    char *s;
    char *k1, *k2;
    char *i1, *i2;
    FILE *wf;
    int  i;
    int  sw;
{
	int pos, pos2, oft, l1, l2;
	char *c1, *c2;
	
	fwrite (s, (size_t)i, (size_t)1, wf);
	c1=strdup(k1);
	if (debug) fprintf (stderr, "Startc1=%s\n", c1);
	oft=0;
	while ((pos=InString(c1, "", oft))!=-1) {
		/*		fprintf(stderr, "pos=%d\n",pos);*/
		if ((pos2=InString(c1, "ʤ", 0))!=-1) {
			c2 = strdup(c1);
			RemoveWord(c2, "");
			RemoveWord(c2, "");
#if defined(Solaris)|| defined(SUN4)|| defined(NESOS6)
                        printf ("%s%s %ld\n", c2, i1, pointer);
#else
			/*printf ("%s%s %p\n", c2, i1, pointer);*/
			printf ("%s%s %ld\n", c2, i1, pointer);
#endif
		}
		l1 = strlen(c1);
		CutStrPart(c1, "", "");
		c1[strlen(c1)+1]='\0';
			if(debug)	fprintf (stderr, "c1=%s\n", c1);
		if (l1==strlen(c1)) {
			break;
		}
		oft = pos;
	}
#if defined(Solaris)|| defined(SUN4) || defined(NEWSOS6)
        printf ("%s%s %ld\n", c1, i1, pointer);
        if (sw!=1) {
                printf ("%s%s %ld\n", k2, i2, pointer);
        }
#else
	/*printf ("%s%s %p\n", c1, i1, pointer);*/
	printf ("%s%s %ld\n", c1, i1, pointer);	  
	if (sw!=1) {
		/*printf ("%s%s %p\n", k2, i2, pointer);*/
		printf ("%s%s %ld\n", k2, i2, pointer);
	}
#endif
	fflush (stdout);
	pointer = ftell(wf);
}

int main (argc, argv)
    int   argc;
    char *argv[];
{
 	FILE *wf;
	char buf[BUFSIZ], elem[20][200], *outd, *org, up[5], mid[5], dw[5];
	int  i, j, k, l;

	pointer = (off_t)0;
	optchk(argc, argv);
	if ((wf=fopen(fn, "w+b"))==NULL) {
		perror ("cannot open wf");
		exit(1);
	}
	while((fgets(buf, BUFSIZ, stdin))!=NULL) {
		i=strlen(buf);
		buf[(i-1)]='\0';
		org = strdup (buf);
		for (j=0; buf[0]!='\0'; j++) {
			if (dic==1||dic==3||dic==4) {
				splitdata (elem[j], buf, ' ');
			} else if(dic==0||dic==2||dic==5){ 
				splitdata (elem[j], buf, ',');
			}
		}
		if((outd=calloc((size_t)i+10, sizeof(char)))==NULL) {
			perror ("calloc failed\n");
			exit(1);
		}
		if (dic==0) { /* newbgh for chainer */
			k=strlen(elem[2]);
			for (l=0; elem[2][l]!='.'; l++) up[l]=elem[2][l];
			up[l]='\0';
			for (j=l+1; j<=k; j++)	dw[(j-l-1)]=elem[2][j];
			dw[(j-l-1)]='\0';
			elem[5][(strlen(elem[5])-1)]='\0';
			sprintf (outd, "%s %s %s,%s,%s,%s,%s",
				 elem[0], elem[1], up,dw,elem[3],elem[4],elem[5]);
			l=strlen(outd)+1;
			mkdata (outd, elem[0], elem[1], "", "R", wf ,l, 1);
		} else if (dic==1) { /* kadokawa for chainer */
			k=strlen(elem[0]);
			for (l=0; l<3; l++) up[l]=elem[0][l];
			up[3]='\0';
			if (k==6) {
				mid[0]='z'; mid[1]='\0';
				for (l=3; l<6; l++) dw[(l-3)]=elem[0][l];
				dw[3]='\0';
			} else if (k==7) {
				mid[0]=elem[0][3]; mid[1]='\0';
				for (l=4; l<7; l++) dw[(l-4)]=elem[0][l];
				dw[3]='\0';
			}
			/*
			sprintf (outd, "%s %s %s,%s,%s,0,0 %s %s %s",
				 elem[3], elem[4], up,mid,dw, 
				 elem[5], elem[1], elem[2]);
		        */
			sprintf (outd, "%s %s %s,%s,%s,0,0 %s",
				 elem[3], elem[4], up,mid,dw, 
				 elem[5]);
			l=strlen(outd)+1;
			mkdata (outd, elem[3], elem[4], "", "R", wf ,l, 1);

		} else if (dic==2) { /* newbgh for nero */
			sprintf (outd, "%s %s %s %s %s %s",
				 elem[2], elem[3], elem[4], elem[5],
				 elem[0], elem[1]);
			if(field==0)
				mkdata(outd,elem[0], elem[1], "", "", wf ,i,1);
			else if(field==1)
				mkdata(outd, elem[0], elem[1], "", "R", wf ,i);
			else if(field==2)
				mkdata(outd, elem[0], elem[1],"W", "R", wf ,i);

			/*mkdata (outd, elem[0], elem[1], "W", "R", wf ,i);*/
			
		} else if (dic==3) { /* kadokawa for nero */
			sprintf (outd, "%s %s %s %s %s %s",
				 elem[0], elem[3], elem[4], elem[5],
				 elem[1], elem[2]);
			if(field==0)
				mkdata(outd,elem[3], elem[4], "", "", wf, i,1);
			else if(field==1)
				mkdata(outd, elem[3], elem[4], "", "R", wf, i);
			else if (field==2)
				mkdata(outd, elem[3], elem[4],"W", "R", wf, i);

		} else if (dic==4) { /* wordnet for nero ..*/
			strcpy(outd, org);
			fprintf (stderr, "%s\n", outd);
			l=strlen(outd)+1;
			mkdata (outd, elem[0], elem[0], "", "R", wf, i, 1);

		} else if (dic==5) { /* ntt thesaurus */
			sprintf (outd, "%s %s %s,%s,%s,0,%s",
				 elem[0], elem[1], elem[2],elem[3],elem[4],elem[5]);
			l=strlen(outd)+1;
			mkdata (outd, elem[0], elem[1], "", "R", wf ,l, 1);
		}

		free(outd); outd=NULL;
		free(org); org=NULL;
	}
	if (dic==0 || dic==1) {
		/* ̤뤿ιԤɲ */
		i=25;
		if((outd=calloc((size_t)i, sizeof(char)))==NULL) {
			perror ("calloc failed\n");
			exit(1);
		}
		if (dic==0) {
			sprintf (outd, "DICID DICID newbgh");
		} else if (dic==1) {
			sprintf (outd, "DICID DICID kadokawa");
		}
		l=strlen(outd)+1;
		mkdata (outd, "DICID", "DICID", "", "R", wf ,l, 1);
		free(outd); outd=NULL;
	}
}
