/*
 * mrp2dat.c
 *
 * Copyright (c) 1999,2000,2001 Hajime Mochizuki <motizuki@jaist.ac.jp>
 *   Natural Language Processing Laboratory,
 *   Japan Advanced Institute of Science and Technology.
 *
 *
 * Use of this software is granted only for RESEARCH PURPOSES. 
 * Redistribution is granted as long as all copyright notices are retained.
 * Redistribution with modification is allowed provided that you make your 
 * modified version obviously distinguishable from the original one.  
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES ARE DISCLAIMED.
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mrp2dat.h"
#include "mrpfil.h"
#include "nero.h"

int  ys = 0;

#define CHA2DAT_OUT_SIZE  (1024*2)
static char *cha2dat_out;
static int cha2dat_out_idx;
static int cha2dat_out_num;
static char datpl[CHA2DAT_OUT_SIZE];

static int dicid=-1; /* 饹̾ */

extern struct patrash_answer_ptr* tell_patrash_ptr();
extern char* search_dicinfo();
extern void splitdata();
extern void free_patrash_answer_ptr();

extern struct part_rl *prl;
extern struct mrp_entry* like_sscanf(); /* mrpfil */
extern void set_mrp_rules ();

static void add_cha2dat (str, output)
    char *str;
    char *output;
{
	int len = strlen(str);
	char *tmp;

	if (cha2dat_out_idx + len >= CHA2DAT_OUT_SIZE * cha2dat_out_num &&
	    cha2dat_out) {
		if((tmp = realloc (cha2dat_out, 
				   CHA2DAT_OUT_SIZE * ++cha2dat_out_num))==NULL) {
			fprintf (stderr, "realloc error\n");
			exit(1);
		}
		cha2dat_out=tmp;
	}

	if (cha2dat_out) {
		strcpy (cha2dat_out + cha2dat_out_idx, str);
		cha2dat_out_idx += len;
	}
}       

void set_cha2dat_out(output)
   FILE *output;
{
	if (output == NULL) {
		if (cha2dat_out_num > 1) {
			free (cha2dat_out); cha2dat_out=NULL;
			cha2dat_out_num = 0;
		} 
		if (cha2dat_out_num == 0) 
			cha2dat_out = malloc(CHA2DAT_OUT_SIZE * ++cha2dat_out_num);
		cha2dat_out_idx = 0;
		cha2dat_out[0] = '\0';
	} else {
		cha2dat_out = (char *)output;
	}
}

struct term_entry *make_c2d (c2d_size)
     int c2d_size;
{
	struct term_entry *tmp;
	int i;

	if ((tmp=calloc(c2d_size,sizeof(*tmp)))==NULL) {
		fprintf (stderr, "calloc for c2d error\n");
		exit(1);
	}
	/*  */
	for (i=0; i<c2d_size; i++) {
		(tmp+i)->term = NULL;
		(tmp+i)->snum = 0;
		(tmp+i)->wnum = 0;
		(tmp+i)->ts_size = 0;
		(tmp+i)->ts_spnd = 0;
		(tmp+i)->ts = NULL;
		(tmp+i)->joshi[0] = '\0';
	}
	return tmp;
}

struct term_sense *cal_ts (p, ts_size)
     struct term_sense *p;
     int ts_size;
{
	int i;
	/*struct term_sense *p;*/
	/*p=NULL;*/

	/*	if ((tmp=calloc((size_t)ts_size,sizeof(*tmp)))==NULL) {*/
	if ((p=calloc(ts_size,sizeof(*p)))
	    ==NULL) {
		fprintf (stderr, "calloc for ts error\n");
		exit(1);
	}
	/*  */
	for (i=0; i<ts_size; i++) {
		/*(p+i)->id = NULL;*/
		(p+i)->yomi = NULL;
	}

	/*
	for (i=0; i<ts_size; i++) {
		if ((t=(char **)calloc((size_t)ss_size, sizeof(**t)))==NULL) {
			fprintf (stderr, "calloc for char error\n");
			exit(1);
		}
		p[i].id=t;
	}
	*/
	return p;
}

struct term_sense *real_ts (ts, ts_size)
     struct term_sense *ts;
     int ts_size;
{
	struct term_sense *tmp;

	/*	if ((tmp=realloc(ts, sizeof(*ts)*ts_size))==NULL) {*/
	if ((tmp=realloc(ts, sizeof(*tmp)*(ts_size)))==NULL) {
		fprintf (stderr, "realloc for ts error\n");
		exit(1);
	}
	return tmp;
}

/*
 * ߻ѤƤ륷饹Ǥ뤫Τ 
 *
 *   -1 .. 顼
 *         0  .. ʬɽ
 *         1  .. 쿷ŵ
 *
 */
int get_thesaurus_id (pa,pwan,dic)
    struct patrash_array *pa;
    FILE *pwan, *dic;
{
	struct patrash_answer_ptr *ptrr;
	char *dicinfo;
	char elem[3][MORPHMAXLEN];
	int i;

	ptrr=tell_patrash_ptr("DICID", pa, pwan); /* ǺϤ*/
	if (ptrr!=NULL) {
		dicinfo=NULL; 
		dicinfo=search_dicinfo (ptrr, dic);
		free_patrash_answer_ptr(ptrr);

		for(i=0; i<3; i++) {
			elem[i][0]='\0';
			splitdata(elem[i], dicinfo, ' ');
		}
		if      (!strcmp(elem[2],"newbgh\n"))   return 0;
		else if (!strcmp(elem[2],"kadokawa\n")) return 1;
		else return 0;
	} else {
		return -1;
	}
}

int dupid_check (ck, id)
     char *ck, *id;
{
	int i;
	char d[IDLEVELMAX][20];
	char one[(IDLEVELMAX*20)], two[(IDLEVELMAX*20)];


	strcpy(one, ck);
	strcpy(two, id);
	if (dicid==1) {/*  */
		d[0][0]='\0';
		d[1][0]='\0';
		splitdata(d[0], one, ',');
		splitdata(d[1], two, ',');
		if (!strcmp(d[0],d[1])) return 1;
		else return 0;
	} else if (dicid==0) {
		for (i=0; i<2; i++) {
			d[i][0]='\0';
			d[(i+2)][0]='\0';
			splitdata(d[i], one, ',');
			splitdata(d[(i+2)], two, ',');
		}
		if (!strcmp(d[0],d[2]) && !strcmp(d[1],d[3])) return 1;
		else return 0;
	} else 
		return 0;
}

/*
 * ignore list
 */
/*int kadokawa_ignore_list (str)*/
int ignore_list (id)
   char *id;
{
	int i;
	char d[IDLEVELMAX][20];
	char one[(IDLEVELMAX*20)];

	strcpy(one, id);
	if (dicid==1) {/*  */
		d[0][0]='\0';
		splitdata(d[0], one, ',');
		if (strcmp("508", d[0])==0) return 1;
		if (strcmp("509", d[0])==0) return 1;
		if (strcmp("828", d[0])==0) return 1;
		if (strcmp("829", d[0])==0) return 1;
		if (strcmp("834", d[0])==0) return 1;
		return 0;
	} else if (dicid==0) { /* ʬɽ (ͽ) */
		return 0;
	}
	return 0;
}

/*
 * kadokawa 
 *
 * Ƥ  c2d ... term_senseʬ⤦äȥޡȤˡ
 *
 * MAXʬ餫tmpȤѰդƤ>
 *         ΰݤ ʬ(ʬ)ˤơΤޤޥԡ
 * Ѥ (reallocɬפʤ褦ˤ)
 * 
 * ȤΤɤ 1999.12.20Ƥ
 *
 */
/*int search_kadokawa(term,snum,wnum,idlevel,pa,pwan,dic)*/
int search_the(term,snum,wnum,idlevel,pa,pwan,dic)
    char *term;
    int  snum, wnum, idlevel;
    struct patrash_array *pa;
    FILE *pwan, *dic;
{
	int i, j, dnum, pf, start, num, ix, omake, k, l, m, n, I;
	/*	char d[KADSENSEMAXNUM][KADSENSEMAXLEN];*/
	char d[15][50];
	char s[3][5], tentity[ELEMMAXLEN], checkid[ELEMMAXLEN];
	struct patrash_answer_ptr *ptrr;
	char *dicinfo, *dici;

	checkid[0]='\0';
	omake=1;

	/*  */
	ptrr=NULL;
	ix = 0;

	if (datpl[0]!='\0') {
		sprintf (datpl, "%s%s\t%d\t%d",datpl, term,  snum, ++wnum);
	} else {
		sprintf (datpl, "%s\t%d\t%d", term,  snum, ++wnum);
	}

	if (idlevel!=0) ptrr=tell_patrash_ptr(term, pa, pwan);
	if (ptrr!=NULL) { /* --- match  ---------------------------*/
		dicinfo=NULL; 
		dicinfo=search_dicinfo (ptrr, dic);
		dici=dicinfo;
		free_patrash_answer_ptr(ptrr);
		for (i=0; dici[0]!='\0'; i++) {
			tentity[0]='\0';
			splitdata(tentity, dici, '\n');
			for (j=0; tentity[0]!='\0'; j++) {
				d[j][0]='\0';
				splitdata (d[j], tentity, ' ');
			}
			/*if(kadokawa_ignore_list(d[3])==0 &&*/
			if(ignore_list(d[2])==0 &&
			   dupid_check(checkid,d[2])==0) {
				if (ix==0)
					sprintf(datpl,"%s\t%s", datpl,d[2]);
				else 
					sprintf(datpl,"%s%s", datpl,d[2]);
				sprintf (datpl, "%s:%s;",datpl,d[1]);
				ix++;
			}
			strcpy(checkid, d[2]);
		}
		free (dicinfo);	dicinfo=NULL;
	}
	/* kadokawa entryä̵ʤΤʤä */
	/* 뤤ϡ no match  ID LEVEL=0ξ */
	if (ix==0){ 
		sprintf (datpl, "%s\t0", datpl);
		for(i=1;i<IDLEVELMAX;i++)
			sprintf (datpl, "%s,0", datpl);
		sprintf (datpl, "%s:%s;", datpl,term);
		return wnum;
	} else {
		return wnum;
	}
}


/*
 * mrp_typeˤäʣηǲϴбǽʤ褦ѹ 27/04/01
 *
 * μΤޤαʤˡ
 *
 *
 */
int cha2dat_the (record, snum, wnum, state, idlevel, pa, pwan, dic, mode, mrp_type)
   char *record;
   int  snum, wnum, state, idlevel;
   struct patrash_array *pa;
   FILE *pwan, *dic;
   int mode, mrp_type;
{
	int  i, j, yf, len, d4, d6, ok, partofspeech;
	char last[3];
	unsigned char tmp[1024];
	struct mrp_entry *me;
	struct mrp_ana *ma;
	char joshitype[10];
	char joshibase[10];

	me=NULL;
	yf = ys;  /* ys = yougen state */
	me = like_sscanf(record, mrp_type);
	len = strlen(me->field[2]);

	/* d4 = ʻʬࡤ d6 = ʬ */
	d4=-1;
	if ((mode<4) && strcmp(me->field[3], prl[PARTNOUN].word)==0) { /* ̾ */
		d4 = 6;
		d6 = 0;
		partofspeech=PARTNOUN;
	} else if ((mode<2||mode>4)&&
		   (strcmp(me->field[3], prl[PARTADJ].word)==0)){/* ƻ */
		d4=3;
		d6=0;
		partofspeech=PARTADJ;
	} else if ((mode<1||(mode>2 && mode<6)) && 
		   (strcmp(me->field[3], prl[PARTVERB].word)==0)) { /* ư */
		d4=2;
		d6=0;
		partofspeech=PARTVERB;
	} else if (strcmp(me->field[3], prl[PARTUNKN].word)==0) { /* ̤ */ 
		d4=6;	/* ̾Ʊ˰ */
		d6=0;
		partofspeech=PARTUNKN;
	} else if (strcmp(me->field[3], prl[PARTPOST].word)==0) { /*  */
		d4=9;
		partofspeech=PARTPOST;
	} else {
		partofspeech=-1;
	}

	if (partofspeech!=-1) {
		if (partofspeech==PARTPOST) {
			strcpy(joshitype, "no");
			strcpy(joshibase, "no");
			for (i=0; i<prl[PARTPOST].rule_ct; i++) {
				ma = (prl+PARTPOST)->ma+i;
				ok = 1;
				for(j=1; j<=ma->fld[0]; j++) {
					if (strcmp(me->field[(ma->fld[j])], 
						   ma->words[j])!=0) {
						j=(ma->fld[0])+1;
						ok=0;
					}
				}
				if (ok) {
					strcpy(joshitype, ma->action);
					strcpy(joshibase, me->field[2]);
					/* 롼ɽ˰̣ */
					i=prl[PARTPOST].rule_ct;
				}
			}
		} else {
			for (i=0; i<prl[partofspeech].rule_ct; i++) {
				ma = (prl+partofspeech)->ma+i;
				ok = 1;
				for(j=1; j<=ma->fld[0]; j++) {
					if (strcmp(me->field[(ma->fld[j])], 
						   ma->words[j])!=0) {
						j=(ma->fld[0])+1;
						ok=0;
					}
				}
				if (ok) {
					d6 = atoi(ma->action); /* -1 Ƚ */
					/* 롼ɽ˰̣ */
					i=prl[partofspeech].rule_ct; 
				}
			}
		}
	}

	/* Ҥ餬1ʸ̾ϥå*/
	if (d4==6 && len==2 && 
	    (me->field[2][0]==0xffffffa4 || me->field[2][0]==0xffffffa5))  {
		if (state==1 && yf==0)	sprintf (datpl, "\tno\tno\n");
		free(me);me=NULL;
		return wnum;
	}

	if (d4==6 && d6==0) {                                   /* for ̾ */
		wnum = search_the(me->field[2], snum, wnum, idlevel, pa, pwan, dic);
		if (state==1 && yf==0) {
			strcpy(tmp,datpl);
			sprintf(datpl,"\tno\tno\n%s", tmp);
		}
		ys=0;
	} else if (d4==3 && d6==0) {                           	/* for ƻ */
		if (state==1 && yf==0)
			sprintf (datpl, "%s\tno\tno\n", datpl);
		j=0; last[0]='\0';
		if (len > 2) {
			for (i=len-2;i<len;i++){
				last[j]=me->field[2][i]; j++;
			}
			last[2]='\0';
			if (strcmp(last, "")==0) me->field[2][len-2]='\0';
		}
		wnum = search_the(me->field[2], snum, wnum, idlevel, pa, pwan, dic);
		sprintf (datpl, "%s\tyougen\tno\n", datpl);
		ys=1;

	} else if (d4==2 && d6==0) {                           /* for ư */
		if (state==1 && yf==0)
			sprintf (datpl, "%s\tno\tno\n", datpl);
		if (d6!=-1) {
			wnum=search_the(me->field[2],snum,wnum,idlevel,pa,pwan,dic);
			sprintf (datpl, "%s\tyougen\tno\n", datpl);
			ys=1;
		} else {
			ys=0;
		}
	} else if (state==1) {
		if (d4==9 && ys==0) { /* for  */
			/*
			if((strcmp(d[5],"ʽ")==0)&&(strcmp(D[0],"")!=0)){
				if (strcmp(D[0], "")==0) 
					sprintf(datpl,"%s\t%s\n",datpl,"gashugo");
				else if (strcmp(D[0], "")==0)
					sprintf(datpl,"%s\t%s\n",datpl,"wotsuki");
				else
					sprintf(datpl,"%s\t%s\n",datpl,"kakujoshi");
			} else if (strcmp(d[5], "")==0) {
				if (strcmp(D[0], "")==0)
					sprintf(datpl,"%s\t%s\n",datpl,"hashugo");
				else if (strcmp(D[0], "")==0)
					sprintf(datpl,"%s\t%s\n",datpl,"motsuki");
				else
					sprintf(datpl,"%s\t%s\n",datpl,"fukujoshi");
			} else if ((strcmp(d[5],"̾³")==0) &&
				   (strcmp(D[0],"")==0)) {
				sprintf (datpl,"%s\t%s\n",datpl,"kakujoshi");
			} else if (yf==0)
				sprintf (datpl,"%s\t%s\n",datpl,"no");
			*/

			/* ѹƤߤ... */
			sprintf (datpl,"%s\t%s\t%s\n",datpl, joshitype, joshibase);
			ys=0;
		} else {
			ys=0;
			if (yf==0) sprintf (datpl,"%s\t%s\t%s\n",datpl,"no", "no");
		}
	} else ys=0;
	free(me); me=NULL;
	return wnum;
}


void testtest()
{
	;
}

int clear_c2d (c2d, c2d_spnd)
     struct term_entry* c2d;
     int c2d_spnd;
{
	int i,j, k;

	for(i=0; i<c2d_spnd; i++) {
		if (c2d[i].ts_size>0) {
			for (j=0; j<c2d[i].ts_spnd; j++) {
				/*
				if (j<c2d[i].ts_spnd) {
					for (k=0; k<IDLEVELMAX; k++) {
						free(c2d[i].ts[j].id[k]);
						c2d[i].ts[j].id[k]=NULL;
					}
				}
				*/
				free(c2d[i].ts[j].yomi);
				c2d[i].ts[j].yomi=NULL;
			}
			free(c2d[i].ts);
			c2d[i].ts=NULL;
			c2d[i].ts_spnd=0;
			c2d[i].ts_size=0;
			free(c2d[i].term);
			c2d[i].term=NULL;
			c2d[i].snum=0;
			c2d[i].wnum=0;
			c2d[i].joshi[0]='\0';
			/*free(c2d[i].joshi);c2d[i].joshi=NULL;*/
		}
	}
	return 0;
}

/*
 * cha_in˷ǲϺѤߤη̤ϤƤ
 */
char* cha2dat (cha_in, snum, allmrp, idlevel, pf, dic, fm, mrp_type, mrf, ciw_sw)
     char *cha_in, *mrf;
     int fm; /* filter mode */
     int  snum, *allmrp, idlevel, mrp_type, ciw_sw;
     struct patrash_files *pf;
     FILE *dic;
{
	int  i, wn, pwn, state, am;
	char cha[BUFSIZ];
	
	wn = state = 0;
	am=*allmrp;
	if (prl==NULL) set_mrp_rules (mrp_type, mrf);

	if (dicid==-1) {
		dicid = get_thesaurus_id (pf->pa,pf->wan,dic);
	}
	set_cha2dat_out(NULL);
	if (cha_in[0]=='\0') 	return NULL;

	for (i=0; cha_in[0]!='\0'; i++) {
		cha[0]='\0';
		splitdata (cha, cha_in, '\n');
		if (strcmp(cha, "EOS")!=0) {
			pwn = wn;
			wn=cha2dat_the(cha, snum, wn, state, idlevel, 
					 pf->pa, pf->wan, dic, fm, mrp_type);
			add_cha2dat (datpl, cha2dat_out);
			datpl[0]='\0';
			if (pwn==wn) { /*ñ줬ͭä?*/
				if (ciw_sw==1) wn++;
				state = 0;
			} else  state = 1;
			am++;
		}
	}
	*allmrp=am;
	if (wn == 0) return NULL;
	else 	     return cha2dat_out;
}

/*
 * mrp_typeˤäʣηǲϴбǽʤ褦ѹ 27/04/01
 */
char* nd_cha2dat (cha_in, snum, wn, idlevel, pf, dic, fm, state, init, mrp_type, mrf)
    char *cha_in, *mrf;
    int  fm; /* filter mode */
    int  snum, *wn, idlevel, state, init, mrp_type;
    struct patrash_files *pf;
    FILE *dic;
{
	int wnum;
	wnum = *wn;

	if (prl==NULL) set_mrp_rules (mrp_type, mrf);
	if (init==1) {
		set_cha2dat_out(NULL);
		if (cha_in[0]=='\0') {
			return NULL;
		}
		if (dicid==-1) {
			dicid = get_thesaurus_id (pf->pa,pf->wan,dic);
		}
	}
	datpl[0]='\0';
	wnum=cha2dat_the(cha_in,snum,wnum,state,idlevel,pf->pa,pf->wan,dic,fm,mrp_type);
	add_cha2dat (datpl, cha2dat_out);
	*wn = wnum;
	return cha2dat_out;
}

/*
 * mrp_typeˤäʣηǲϴбǽʤ褦ѹ 27/04/01
 */
char* nd_cha2dat_old (cha_in, snum, wn, idlevel, pa, pwan, dic, fm, state, init, mrp_type, mrf)
    char *cha_in;
    int  snum, *wn, idlevel;
    struct patrash_array *pa;
    FILE *pwan, *dic;
    int fm; /* filter mode */
    int state, init, mrp_type;
    char *mrf;
{
	int wnum;
	wnum = *wn;

	if (prl==NULL) set_mrp_rules (mrp_type, mrf);
	if (init==1) {
		set_cha2dat_out(NULL);
		if (cha_in[0]=='\0') {
			return NULL;
		}
		if (dicid==-1) {
			dicid = get_thesaurus_id (pa,pwan,dic);
		}
	}
	datpl[0]='\0';
	wnum = cha2dat_the(cha_in, snum,wnum, state,idlevel, pa,pwan,dic, fm, mrp_type);
	add_cha2dat (datpl, cha2dat_out);
	*wn = wnum;
	return cha2dat_out;
}
